'use strict'

const { constants } = require('fs')
const fs = require('fs/promises')
const sharp = require('sharp')
const path = require('path')

/**
 * @readonly
 */
const REG_ONLYIMG = /\.png|\.jpg/g
const REG_GETPATH = /src|=|"|\s|from|'/g
const REG_MATCHIMG = /src\=(\"|\').*\.(png|jpg)(\"|\')|from.*(\"|\').*\.(png|jpg)(\"|\')/g

/**
 * no line break console.
 * 
 * @param {string} msg
 * @returns {void}
 * @private
 */
function print(msg) {
    process.stdout.write(msg)
}

/**
 * convert image path to webp
 * 
 * @param {string} url
 * @returns {string}
 * @private
 */
function convertPath(imagePath) {
    const { dir, root, name } = path.parse(imagePath)
    return path.format({ dir, root, name, ext: '.webp' })
}

/**
 * file exists.
 * 
 * @param {string} path
 * @returns {Promise<bool>}
 * @private
 */
function fsExists(path) {
    fs.access(path, constants.R_OK | constants.W_OK)
        .then(() => true)
        .catch(() => false)
}

/**
 * compress image.
 * jpg and png to webp.
 * 
 * @param {string[]} files
 * @returns {Promise<string[]>}
 * @private
 */
async function compressImage(files) {
    let outputs = []
    for (const file of files) {
        const outPath = convertPath(file)
        
        // Do nothing if the webp file already exists。
        if (await fsExists(outPath)) {
            continue
        }

        await sharp(file).webp().toFile(outPath)
        print(`\ncompress: ${file} -> WEBP`)
        outputs.push(outPath)
    }
    
    /**
     * The generated webp files are recorded here in order to delete these files after 
     * the build is completed, because these files will affect git diff.
     */
    return outputs
}

module.exports = function() {
    let alias = null
    let outputs = []
    
    return {
        name: 'vite:compress',
        apply: 'build',
        configResolved(config) {
            /**
             * Compatible with custom modules to introduce address 
             * rewriting and restore the real path。
             */
            alias = config.resolve.alias
        },
        async transform(code, filePath) {
            if (code.match(REG_ONLYIMG)) {
                let files = []
                const replacedCode = code.replace(REG_MATCHIMG, key => {
                    files.push(alias.reduce((next, item) => {
                        return next.match(item.find) ? 
                            next.replace(item.find, item.replacement) : 
                            next
                    }, key.replace(REG_GETPATH, '')))
                    return key.replace(REG_ONLYIMG, '.webp')
                })
                
                outputs.push(...await compressImage(files))
                return replacedCode
            } else {
                return code
            }
        },
        async closeBundle() {
            for (const file of outputs) {
                if (await fsExists(file)) {
                    await fs.rm(file)
                }
            }
        }
    }
}
